compute_workflow <- function(recipe, hcst, obs, fcst = NULL) {
  modules <- tolower(strsplit(recipe$Run$startR_workflow$modules,
                              ", | |,")[[1]])
  # Create data list
  data <- list(hcst = list(), obs = list(), fcst = list())
  # Fill list following the structure of the s2dv_cube
  data$hcst$data <- hcst
  data$obs$data <- obs
  data$fcst$data <- fcst
  # Remove duplicated objects
  rm(hcst, obs, fcst)
  # Pre-process datasets
  data <- preprocess_datasets(recipe, data)
  # Define other outputs
  skill_metrics <- NULL
  # Loop over the modules
  for (module in modules) {
    if (module == "calibration") {
      data <- calibrate_datasets(recipe, data)
    } else if (module == "anomalies") {
      data <- compute_anomalies(recipe, data)
    } else if (module == "skill") {
      skill_metrics <- compute_skill_metrics(recipe, data)
    }
  }
  ## TODO: Define what to return depending the modules called + the recipe
  return_list <- list(hcst = data$hcst$data,
                      obs = data$obs$data,
                      fcst = data$fcst$data,
                      skill = skill_metrics)
  # Eliminate NULL elements from the return list
  return_list <- return_list[!sapply(return_list, is.null)]
  return(return_list)
}

convert_to_s2dv_cube <- function(new_cube, original_cube) {
  ## TODO: Make sure all the metadata is created correctly
  attr(new_cube, "Variables") <- attr(original_cube, "Variables")
  attr(new_cube, "FileSelectors") <- attr(original_cube, "FileSelectors")

  class(new_cube) <- "startR_array"
  new_cube <- as.s2dv_cube(new_cube)
  return(new_cube)
}

run_compute_workflow <- function(recipe, data) {
  
  # ---------------------------------------------------------------------------
  # Step 1: Retrieve the modules that will be called inside the workflow

  modules <- tolower(strsplit(recipe$Run$startR_workflow$modules,
                              ", | |,")[[1]])
  
  # ---------------------------------------------------------------------------
  # Step 2: Define the inputs and outputs for Compute()
  # Retrieve original dimensions
  exp_dims <- names(attr(data$hcst, "Dimensions"))
  obs_dims <- names(attr(data$obs, "Dimensions"))
  # Remove the chunking dimensions from the original dimensions in the arrays,
  # to create the target dimensions
  exp_target_dims <- 
    exp_dims[!exp_dims %in% names(recipe$Run$startR_workflow$chunk_along)]
  obs_target_dims <-
    obs_dims[!obs_dims %in% names(recipe$Run$startR_workflow$chunk_along)]
  # Default inputs: hindcast and observations
  target_dims <- list(hcst = exp_target_dims,
                      obs = obs_target_dims)
  inputs <- list(hcst = data$hcst, obs = data$obs)
  # Add forecast if not empty
  if (!is.null(data$fcst)) {
    target_dims <- c(target_dims,
                     list(exp_target_dims))
    inputs <- c(inputs, list(fcst = data$fcst))
  }

  # Create output dimensions
  ## TODO: Add several conditions depending on module/user requests?
  ## For example: if 'downscaling' or 'Indices' in modules, lat and lon cannot
  ## be used as chunking dimensions
  exp_output_dims <- exp_target_dims[!exp_target_dims == 'var_dir']
  # Default outputs: hindcast and observations
  output_dims <- list(hcst = exp_output_dims,
                      obs = exp_output_dims)
  # Add forecast if not empty
  if (!is.null(data$fcst)) {
    output_dims <- c(output_dims,
                     list(fcst = exp_output_dims))
  }
  # Add skill metrics if skill module is called
  if ("skill" %in% modules) {
    skill_dims <- c('metric', 'var', 'time', 'latitude', 'longitude')
    skill_output_dims <-
      skill_dims[!skill_dims %in% names(recipe$Run$startR_workflow$chunk_along)]
    output_dims <- c(output_dims,
                     list(skill = skill_output_dims))
  }

  # ---------------------------------------------------------------------------
  # Step 3: Generate the Step and call Compute()
  ## TODO: Add use attributes
  step <- Step(fun = compute_workflow,
               target_dims = target_dims,
               output_dims = output_dims)
  wf <- AddStep(inputs = inputs,
                step = step,
                recipe = recipe)
  res <- Compute(wf$hcst,
                 chunks = recipe$Run$startR_workflow$chunk_along)

  browser()
  info(recipe$Run$logger,
       "##### COMPUTE SECTION ENDED, REFORMATTING OUTPUT #####")

  # ---------------------------------------------------------------------------
  # Step 4: Convert results to s2dv_cube/list format
  result <- list()
  for (cube in c("hcst", "obs", "fcst")) {
    if (!is.null(res[[cube]])) {
      res$data[[cube]] <- convert_to_s2dv_cube(res[[cube]], data[[cube]])
    } else {
      res$data[[cube]] <- NULL
    }
  }
  ## TODO: Transform skill metrics back into a NAMED list. To do this, Apply()
  ## should return the attributes of the skill array so that we have the name
  ## of each metric, or we should find an alternative way to retrieve them.
  if (!is.null(res$skill)) {
    metric_list <- c("rpss", "rpss_significance") ## to be removed
    res$skill <- lapply(seq(dim(res$skill)[1]), function(x) {res$skill[x, , , , , drop = F]})
    ## TODO: Subset 'metric' dimension
    names(res$skill) <- metric_list
  }
  info(recipe$Run$logger,
       "##### DATA RETURNED AS A NAMED LIST #####")
  return(res)
}
